@@ -22,6 +22,8 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <unistd.h>
+#include <assert.h>
+#include <errno.h>
 
 #include "sigcx/dispatch.h"
 
@@ -164,7 +166,7 @@ StandardDispatcher::StandardDispatcher()
 
 StandardDispatcher::TimerEvent::TimerEvent(const Handler& _cb,
                                            const TimeVal& tmout)
-    : cb(_cb)
+    : cb(_cb), deleted(false)
 {
   expiration.get_current_time();
 
@@ -181,6 +183,7 @@ StandardDispatcher::add_input_handler(const Handler& cb, int fd)
   Guard guard(mutex_);
 
   HandlerID id = get_new_handler_id();
+  ++rd_fd_use_[fd];
   FD_SET(fd, &rd_fds_);
   fd_handlers_.insert(make_pair(id, FileEvent(cb, Read, fd)));
   
@@ -193,6 +196,7 @@ StandardDispatcher::add_output_handler(const Handler& cb, int fd)
   Guard guard(mutex_);
 
   HandlerID id = get_new_handler_id();
+  ++wr_fd_use_[fd];
   FD_SET(fd, &wr_fds_);
   fd_handlers_.insert(make_pair(id, FileEvent(cb, Write, fd)));
   
@@ -204,6 +208,7 @@ StandardDispatcher::add_exception_handler(const Handler& cb, int fd)
 {
   Guard guard(mutex_);
   HandlerID id = get_new_handler_id();
+  ++ex_fd_use_[fd];
   FD_SET(fd, &ex_fds_);
   fd_handlers_.insert(make_pair(id, FileEvent(cb, Except, fd)));
   
@@ -227,6 +232,42 @@ StandardDispatcher::add_timeout_handler(const Handler& cb,
 void StandardDispatcher::remove(HandlerID id)
 {
   Guard guard(mutex_);
+
+  // Queue the removal as to not damage the iterators if we're being called by
+  // a callback in run().
+
+  TMHandlerMap::iterator tm_it = tm_handlers_.find(id);
+  if (tm_it != tm_handlers_.end())
+  {
+    TMEventMap::iterator ev_it = tm_it->second;
+    ev_it->first.deleted = true;
+    return;
+  }
+
+  FDHandlerMap::iterator fd_it = fd_handlers_.find(id);
+  if (fd_it != fd_handlers_.end())
+  {
+    FileEvent& fevent = fd_it->second;
+    
+    if (fevent.ev == Read && --rd_fd_use_[fevent.fd] == 0) {
+      FD_CLR(fevent.fd, &rd_fds_);
+      rd_fd_use_.erase(fevent.fd);
+    }
+    if (fevent.ev == Write && --wr_fd_use_[fevent.fd] == 0) {
+      FD_CLR(fevent.fd, &wr_fds_);
+      wr_fd_use_.erase(fevent.fd);
+    }
+    if (fevent.ev == Except && --ex_fd_use_[fevent.fd] == 0) {
+      FD_CLR(fevent.fd, &ex_fds_);
+      ex_fd_use_.erase(fevent.fd);
+    }
+    
+    fevent.deleted = true;
+  }
+}
+
+void StandardDispatcher::real_remove(HandlerID id)
+{
   
   TMHandlerMap::iterator tm_it = tm_handlers_.find(id);
   if (tm_it != tm_handlers_.end())
@@ -240,15 +281,6 @@ void StandardDispatcher::remove(HandlerID id)
   FDHandlerMap::iterator fd_it = fd_handlers_.find(id);
   if (fd_it != fd_handlers_.end())
   {
-    FileEvent& fevent = fd_it->second;
-    
-    if (fevent.ev == Read && FD_ISSET(fevent.fd, &rd_fds_))
-      FD_CLR(fevent.fd, &rd_fds_);
-    if (fevent.ev == Write && FD_ISSET(fevent.fd, &wr_fds_))
-      FD_CLR(fevent.fd, &wr_fds_);
-    if (fevent.ev == Except && FD_ISSET(fevent.fd, &ex_fds_))
-      FD_CLR(fevent.fd, &ex_fds_);
-    
     fd_handlers_.erase(fd_it);
     return;
   }
@@ -269,10 +301,6 @@ bool StandardDispatcher::run(bool infinite)
     // We set it here, it may be unset in a handler by calling exit
     do_exit_ = false;
 
-    fd_set rd = rd_fds_;
-    fd_set wr = wr_fds_;
-    fd_set ex = ex_fds_;
-    
     // We service timeouts in two passes: first collect all expired
     // timeouts, then invoke them. If we did this in one pass, the
     // callbacks could starve the FD handlers by adding new timeouts
@@ -300,7 +328,12 @@ bool StandardDispatcher::run(bool infinite)
       try
       {
         UnGuard unguard(mutex_);
-        const_cast<Handler&>(ev_it->first.cb)();
+        // Only call it if it isn't queued to be removed - a timeout function
+        // could have tried to remove the one we're looking at earlier in this
+        // pass.
+        if (!ev_it->first.deleted) {
+          const_cast<Handler&>(ev_it->first.cb)();
+        }
       }
       catch (...)
       {
@@ -317,29 +350,43 @@ bool StandardDispatcher::run(bool infinite)
     if (do_exit_)
       break;
     
-    if (tm_events_.size() == 0)
+    // Initialize the fd sets here, as a timeout callback could have added a
+    // new FileEvent.  Otherwise, we'd not be able to wake up from select as
+    // the fd wouldn't be in the sets.
+    fd_set rd = rd_fds_;
+    fd_set wr = wr_fds_;
+    fd_set ex = ex_fds_;
+    
+    // Loop on select, restarting if we were interrupted by a signal.
+    int result;
+    do
     {
-      UnGuard unguard (mutex_);
-      select(FD_SETSIZE, &rd, &wr, &ex, 0);
-    }
-    else
-    {
-      now.get_current_time();
-      timeout = tm_events_.begin()->first.expiration - now;
+      if (tm_events_.size() == 0)
+      {
+        UnGuard unguard (mutex_);
+        result = select(FD_SETSIZE, &rd, &wr, &ex, 0);
+        assert(result != -1 || errno == EINTR);
+      }
+      else
+      {
+        now.get_current_time();
+        timeout = tm_events_.begin()->first.expiration - now;
 
-      UnGuard unguard(mutex_);
-      
-      if (timeout < TimeVal(0))
-        timeout = TimeVal(0);
-      
-      struct timeval tv;
-      tv.tv_sec = timeout.tv_sec;
-      tv.tv_usec = timeout.tv_usec;
-      
-      select(FD_SETSIZE, &rd, &wr, &ex, &tv);
-    }
+        UnGuard unguard(mutex_);
+        
+        if (timeout < TimeVal(0))
+          timeout = TimeVal(0);
+        
+        struct timeval tv;
+        tv.tv_sec = timeout.tv_sec;
+        tv.tv_usec = timeout.tv_usec;
+        
+        result = select(FD_SETSIZE, &rd, &wr, &ex, &tv);
+        assert(result != -1 || errno == EINTR);
+      }
+    } while(result == -1 && errno == EINTR);
 
-    // check after select, we might have caught a signal
+    // check if we've been asked to exit.  If so, exit early.
     if (do_exit_)
       break;
 
@@ -360,11 +407,25 @@ bool StandardDispatcher::run(bool infinite)
       if (event != All)
       {
         UnGuard unguard(mutex_);
-        fevent.cb();
+        // Only call it if it isn't queued to be removed - a fevent function
+        // could have tried to remove the one we're looking at earlier.
+        if (!fevent.deleted)
+          fevent.cb();
       }
       if (do_exit_)
         break;
     }
+
+    // Delete any file events that are marked deleted.
+    for (j = fd_handlers_.begin(); j != fd_handlers_.end(); )
+    {
+      FileEvent& fevent = j->second;
+      HandlerID id = j->first;
+      ++j;
+      if (fevent.deleted)
+        real_remove(id);
+    }
+
   } while (infinite && !do_exit_);
   
   return do_exit_;
